/*
    DeaDBeeF - ultimate music player for GNU/Linux systems with X11
    Copyright (C) 2009-2010 Alexey Yakovenko <waker@users.sourceforge.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include <dbus/dbus.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include "mpris.h"
#include "../../deadbeef.h"
#include "../../config.h"	// need VERSION

DB_functions_t *deadbeef;
DB_misc_t plugin;

int dbus_loop = 1;
int db_conf[3];		// [0]:playback.order, [1]:playback.loop, [2]:cap flag


// put track's metadata info into ret
void mpris_metadata(DBusMessageIter *ret, DB_playItem_t *track) {
	DBusMessageIter dict;
	DBusMessageIter entry;
	char *key;
	DBusMessageIter value;
	dbus_uint32_t v_u32;
	char v_str[32768];
	char *v_str_p=v_str;
	int i;

	mpris_meta metas[]={
		{"location", DBUS_TYPE_STRING, "%F"},
		{"title", DBUS_TYPE_STRING, "%t"},
		{"artist", DBUS_TYPE_STRING, "%a"},
		{"album", DBUS_TYPE_STRING, "%b"},
		{"tracknumber", DBUS_TYPE_STRING, "%n"},
		{"time", DBUS_TYPE_UINT32, "%l"},
		{"mtime", DBUS_TYPE_UINT32, "%l"},
		{"genre", DBUS_TYPE_STRING, "%g"},
		{"comment", DBUS_TYPE_STRING, "%c"},
//		{"rating", DBUS_TYPE_UINT32, "%??"},
		{"year", DBUS_TYPE_UINT32, "%y"},
//		{"date", DBUS_TYPE_UINT32, "%y"},
//		{"arturl", DBUS_TYPE_STRING, "%??"},
//		{"asin", DBUS_TYPE_STRING, "%??"},
//		{"puid fingerprint", DBUS_TYPE_STRING, "%??"},
//		{"mb track id", DBUS_TYPE_STRING, "%??"},
//		{"mb artist id", DBUS_TYPE_STRING, "%??"},
//		{"mb artist sort name", DBUS_TYPE_STRING, "%??"},
//		{"mb album id", DBUS_TYPE_STRING, "%??"},
//		{"mb release date", DBUS_TYPE_STRING, "%??"},
//		{"mb album artist", DBUS_TYPE_STRING, "%??"},
//		{"mb album artist id", DBUS_TYPE_STRING, "%??"},
//		{"mb album artist sort name", DBUS_TYPE_STRING, "%??"},
//		{"audio-bitrate", DBUS_TYPE_UINT32, "%??"},
//		{"audio-samplerate", DBUS_TYPE_UINT32, "%??"},
//		{"video-bitrate", DBUS_TYPE_UINT32, "%??"},
	};

	dbus_message_iter_open_container(ret, DBUS_TYPE_ARRAY, "{sv}", &dict);
		for (i = 0; i < sizeof(metas)/sizeof(mpris_meta); i++) {
			dbus_message_iter_open_container(&dict, DBUS_TYPE_DICT_ENTRY, NULL, &entry);
				dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &metas[i].tag);
				if(metas[i].type == DBUS_TYPE_STRING) {
					dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT, "s", &value);
						if (track) {
							if(!strcmp(metas[i].tag, "location")) {
								deadbeef->pl_format_title(track, -1, v_str, sizeof v_str, -1, metas[i].db_tag);
								if (v_str[0] == '/') {
									strncpy(v_str, "file://", sizeof v_str);
									deadbeef->pl_format_title(track, -1, v_str + 7, sizeof v_str - 7, -1, metas[i].db_tag);
								}
							} else {
								deadbeef->pl_format_title(track, -1, v_str, sizeof v_str, -1, metas[i].db_tag);
							}
						} else {
							v_str[0] = 0;
						}
						dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &v_str_p);
					dbus_message_iter_close_container(&entry, &value);
				} else {
					dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT, "u", &value);
						if (track) {
							if (!strcmp(metas[i].tag, "time")) {
								v_u32 = deadbeef->pl_get_item_duration(track);
							} else if (!strcmp(metas[i].tag, "mtime")) {
								v_u32 = deadbeef->pl_get_item_duration(track) * 1000;
							} else {
								deadbeef->pl_format_title(track, -1, v_str, sizeof v_str, -1, metas[i].db_tag);
								v_u32 = atoi(v_str);
							}
						} else {
							v_u32 = 0;
						}
						dbus_message_iter_append_basic(&value, DBUS_TYPE_UINT32, &v_u32);
					dbus_message_iter_close_container(&entry, &value);
				}
			dbus_message_iter_close_container(&dict, &entry);
		}
	dbus_message_iter_close_container(ret, &dict);
}


// put player's status info into ret
void mpris_status(DBusMessageIter *ret, int evt) {
	DBusMessageIter ret_struct;
	dbus_int32_t a_int32;

	dbus_message_iter_open_container(ret, DBUS_TYPE_STRUCT, NULL, &ret_struct);

		// Playback_State: 0 = Playing, 1 = Paused, 2 = Stopped. 
		// even on DB_EV_SONGSTARTED/DB_EV_SONGFINISHED, deadbeef->get_output()->state() returns OUTPUT_STATE_STOPPED/OUTPUT_STATE_PLAYING
		if (evt == DB_EV_SONGSTARTED || evt == DB_EV_SONGFINISHED) {
			if (evt == DB_EV_SONGSTARTED) {
				a_int32 = 0;
			} else {
				a_int32 = 2;
			}
		} else {
			struct DB_output_s *output = deadbeef->get_output();
			int pl_state = output->state();
			switch (pl_state) {
				case OUTPUT_STATE_STOPPED:
					a_int32 = 2;break;
				case OUTPUT_STATE_PLAYING:
					a_int32 = 0;break;
				case OUTPUT_STATE_PAUSED:
					a_int32 = 1;break;
			}
		}
		dbus_message_iter_append_basic(&ret_struct, DBUS_TYPE_INT32, &a_int32);

		// Shuffle_State: 0 = Playing linearly, 1 = Playing randomly. 
		int pl_order = deadbeef->conf_get_int ("playback.order", 0);
		if (pl_order == PLAYBACK_ORDER_LINEAR) {
			a_int32 = 0;
		} else {
			a_int32 = 1;
		}
		dbus_message_iter_append_basic(&ret_struct, DBUS_TYPE_INT32, &a_int32);

		int pl_loop_mode = deadbeef->conf_get_int ("playback.loop", 0);
		// Repeat_Current_State: 0 = Go to the next element once the current has finished playing, 1 = Repeat the current element.
		if (pl_loop_mode == PLAYBACK_MODE_LOOP_SINGLE) {
			a_int32 = 1;
		} else {
			a_int32 = 0;
		}
		dbus_message_iter_append_basic(&ret_struct, DBUS_TYPE_INT32, &a_int32);

		// Endless_State: 0 = Stop playing once the last element has been played, 1 = Never give up playing 
		if (pl_loop_mode == PLAYBACK_MODE_LOOP_ALL) {
			a_int32 = 1;
		} else {
			a_int32 = 0;
		}
		dbus_message_iter_append_basic(&ret_struct, DBUS_TYPE_INT32, &a_int32);

	dbus_message_iter_close_container(ret, &ret_struct);
}

// return current capabilities
int mpris_caps(int evt) {
	int cap = 0;
	int count = 0;
	DB_playItem_t *track, *prev, *next;
	struct DB_output_s* output;

	track = deadbeef->streamer_get_playing_track();
	count = deadbeef->pl_getcount(PL_MAIN);

	// CAN_GO_NEXT = 1 << 0
    // There is a current next track, or at least something that equals to it (that is, the remote can call the 'Next' method on the interface, and expect something to happen, heh). 
	if (!track) {
		if (count > 0) {
			cap |= 1 << 0;
		}
	} else {
		next = deadbeef->pl_get_next(track, PL_MAIN);
		if (next) {
			cap |= 1 << 0;
			deadbeef->pl_item_unref(next);
		}
	}

	// CAN_GO_PREV = 1 << 1
    // Same as for NEXT, just previous track/something. 
	if (!track) {
		if (count > 0) {
			cap |= 1 << 1;
		}
	} else {
		prev = deadbeef->pl_get_prev(track, PL_MAIN);
		if (prev) {
			cap |= 1 << 1;
			deadbeef->pl_item_unref(prev);
		}
	}

	// CAN_PAUSE = 1 << 2
    // Can currently pause. This might not always be possible, and is yet another hint for frontends as to what to indicate. 
	if (track) {
		// even on DB_EV_SONGSTARTED/DB_EV_SONGFINISHED, deadbeef->get_output()->state() returns OUTPUT_STATE_STOPPED/OUTPUT_STATE_PLAYING
		if(evt == DB_EV_SONGSTARTED || evt == DB_EV_SONGFINISHED) {
			if(evt == DB_EV_SONGSTARTED) {
				cap |= 1 << 2;
			}
		} else {
			output = deadbeef->get_output();
			if (OUTPUT_STATE_STOPPED != output->state()) {
				cap |= 1 << 2;
			}
		}
	}
	
	// CAN_PLAY = 1 << 3
    // Whether playback can currently be started. This might not be the case if e.g. the playlist is empty in a player, or similar conditions. Here, again, it is entirely up to the player to decide when it can play or not, and it should signalize this using the caps API. 
	if (count > 0) {
		cap |= 1 << 3;
	}

	// CAN_SEEK = 1 << 4
    // Whether seeking is possible with the currently played stream (UIs/frontends can then enable/disable seeking controls). 
	if (track) {
		cap |= 1 << 4;
	}
	
	// CAN_PROVIDE_METADATA = 1 << 5
    // Whether metadata can be acquired for the currently played stream/source using GetMetadata at all. 
	if (track) {
		cap |= 1 << 5;
	}
	
	// CAN_HAS_TRACKLIST = 1 << 6
    // Whether the media player can hold a list of several items. 
	cap |= 1 << 6;

	if (track) {
		deadbeef->pl_item_unref(track);
	}

	return cap;
}

void *mpris(void *arg) {
	DBusError err;
	DBusConnection *conn;
	DBusMessage* msg;
	DBusMessage* reply;
	DBusMessageIter args;
	DBusMessageIter ret;
	DBusMessageIter subret;

	DB_playItem_t *track;

	dbus_uint16_t a_uint16;
	dbus_int32_t a_int32;
	dbus_bool_t a_bool;

	int result;
	
	// initialise the error
	dbus_error_init(&err);

	// connect to the bus and check for errors
	conn = dbus_bus_get(DBUS_BUS_SESSION, &err);
	if (dbus_error_is_set(&err)) {
		fprintf(stderr, "mpris: Connection Error (%s)\n", err.message); 
		dbus_error_free(&err); 
		pthread_exit(NULL);
		return NULL;
	}
	if (NULL == conn) {
		fprintf(stderr, "mpris: Connection Null\n"); 
		dbus_error_free(&err); 
		pthread_exit(NULL);
		return NULL;
	}

	// request our name on the bus and check for errors
	result = dbus_bus_request_name(conn, MPRIS_DBUS_NAME, DBUS_NAME_FLAG_DO_NOT_QUEUE | DBUS_NAME_FLAG_REPLACE_EXISTING, &err);
	if (dbus_error_is_set(&err)) {
		fprintf(stderr, "mpris: Name Error (%s)\n", err.message);
		dbus_error_free(&err);
		pthread_exit(NULL);
		return NULL;
	}
	if (DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER != result) {
		fprintf(stderr, "mpris: Not Primary Owner (%d)\n", result);
		dbus_error_free(&err);
		pthread_exit(NULL);
		return NULL;
	}

	while (dbus_loop) {
		dbus_connection_read_write(conn, 10);
    	msg = dbus_connection_pop_message(conn);
		
		// loop again if we haven't got a message
		if (NULL == msg) {
			usleep (10);
			continue; 
		}
      
		#if 0
			printf("mpris: DEST:%s\n",dbus_message_get_destination(msg));
			printf("mpris: PATH:%s\n",dbus_message_get_path(msg));
			printf("mpris: INTERFACE:%s\n",dbus_message_get_interface(msg));
			printf("mpris: MEMBER:%s\n",dbus_message_get_member(msg));
		#endif

		reply = dbus_message_new_method_return(msg);
		dbus_message_iter_init_append(reply, &ret);

		if (dbus_message_is_method_call(msg, "org.freedesktop.DBus.Introspectable", "Introspect")) {
			if (dbus_message_has_path(msg, MPRIS_PATH_ROOT)) {
				dbus_message_iter_append_basic(&ret, DBUS_TYPE_STRING, &MPRIS_ROOT_XML);
			} else if (dbus_message_has_path(msg, MPRIS_PATH_PLAYER)) {
				dbus_message_iter_append_basic(&ret, DBUS_TYPE_STRING, &MPRIS_PLAYER_XML);
			} else if (dbus_message_has_path(msg, MPRIS_PATH_TRACKLIST)) {
				dbus_message_iter_append_basic(&ret, DBUS_TYPE_STRING, &MPRIS_TRACKLIST_XML);
			}
		} else if (dbus_message_has_interface(msg, MPRIS_INTERFACE)) {
			if (dbus_message_has_path(msg, MPRIS_PATH_ROOT)) {
				if (dbus_message_has_member(msg, "Identity")) {
					// Returns a string containing the media player identification. 
					const char *ver = "deadbeef "VERSION;
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_STRING, &ver);
				} else if (dbus_message_has_member(msg, "MprisVersion")) {
					// Returns a struct that represents the version of the MPRIS spec being implemented.
					dbus_message_iter_open_container(&ret, DBUS_TYPE_STRUCT, NULL, &subret);
					a_uint16 = 1;
					dbus_message_iter_append_basic(&subret, DBUS_TYPE_UINT16, &a_uint16);
					a_uint16 = 0;
					dbus_message_iter_append_basic(&subret, DBUS_TYPE_UINT16, &a_uint16);
					dbus_message_iter_close_container(&ret, &subret);
				} else if (dbus_message_has_member(msg, "Quit")) {
					// Makes the "Media Player" exit. 
					deadbeef->quit();
				}
			} else if (dbus_message_has_path(msg, MPRIS_PATH_PLAYER)) {
				if (dbus_message_has_member(msg, "Next")) {
					// Goes to the next element. (What if we're at the end? NOTE Nothing terrible needs to happen,
					// the player should just ignore it. However UIs, and maybe not only UIs, can receive a hint 
					// as to whether there is a next track using the Caps API)
					track = deadbeef->streamer_get_playing_track();
					if(!track) {
						deadbeef->playback_next();
					} else {
						DB_playItem_t *next_track = deadbeef->pl_get_next(track, PL_MAIN);
						if (next_track) {
							// not at end
							deadbeef->playback_next();
							deadbeef->pl_item_unref(next_track);
						}
						deadbeef->pl_item_unref(track);
					}
				} else if (dbus_message_has_member(msg, "Prev")) {
					// Goes to the previous element.(what if we're at the beginning? See note for Next)
					track = deadbeef->streamer_get_playing_track();
					if(!track) {
						deadbeef->playback_prev();
					} else {
						DB_playItem_t *prev_track = deadbeef->pl_get_prev(track, PL_MAIN);
						if (prev_track) {
							// not at beginning
							deadbeef->playback_prev();
							deadbeef->pl_item_unref(prev_track);
						}
						deadbeef->pl_item_unref(track);
					}
				} else if (dbus_message_has_member(msg, "Pause")) {
					// If playing : pause. If paused : unpause. 
					deadbeef->playback_pause();
				} else if (dbus_message_has_member(msg, "Stop")) {
					// Stop playing. 
					deadbeef->playback_stop();
				} else if (dbus_message_has_member(msg, "Play")) {
					// If playing : rewind to the beginning of current track, else: start playing. 
					deadbeef->playback_play();
				} else if (dbus_message_has_member(msg, "Repeat")) {
					// Toggle the current track repeat. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_BOOLEAN != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not boolean!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_bool);
						deadbeef->conf_set_int ("playback.loop", a_bool?PLAYBACK_MODE_LOOP_SINGLE:PLAYBACK_MODE_NOLOOP);
						deadbeef->sendmessage(M_CONFIGCHANGED, 0, 0, 0);
					}
				} else if (dbus_message_has_member(msg, "GetStatus")) {
					// Return the status of media player as a struct of 4 ints. 
					mpris_status(&ret, -1);
				} else if (dbus_message_has_member(msg, "GetMetadata")) {
					// Gives all meta data available for the currently played element.
					// Guidelines for field names are at http://wiki.xmms2.xmms.se/wiki/MPRIS_Metadata
					track = deadbeef->streamer_get_playing_track();
					mpris_metadata(&ret, track);
					if (track) {
						deadbeef->pl_item_unref(track);
					}
				} else if (dbus_message_has_member(msg, "GetCaps")) {
					// Return the "media player"'s current capabilities.
					a_int32 = mpris_caps(-1);
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_INT32, &a_int32);
				} else if (dbus_message_has_member(msg, "VolumeSet")) {
					// Sets the volume. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_INT32 != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not Int32!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_int32);
						deadbeef->volume_set_amp(((float)a_int32)/100);
					}
				} else if (dbus_message_has_member(msg, "VolumeGet")) {
					// Returns the current volume. Volume type: integer between [0;100].
					a_int32 = deadbeef->volume_get_amp() * 100;
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_INT32, &a_int32);
				} else if (dbus_message_has_member(msg, "PositionSet")) {
					// Sets the playing position. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_INT32 != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not Int32!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_int32);
						track = deadbeef->streamer_get_playing_track();
						if(track) {
							float dur = deadbeef->pl_get_item_duration(track);
							if(a_int32 > dur * 1000) {
								deadbeef->streamer_seek(dur);
							} else {
								deadbeef->streamer_seek(((float)a_int32)/1000.0f);
							}
							deadbeef->pl_item_unref(track);
						}
					}
				} else if (dbus_message_has_member(msg, "PositionGet")) {
					// Returns the current playing position. in ms
					a_int32 = deadbeef->streamer_get_playpos() * 1000;
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_INT32, &a_int32);
				}
			} else if (dbus_message_has_path(msg, MPRIS_PATH_TRACKLIST)) {
				if (dbus_message_has_member(msg, "GetMetadata")) {
					// Gives all meta data available for element at given position in the TrackList, counting from 0.
					// Guidelines for field names are at http://wiki.xmms2.xmms.se/wiki/MPRIS_Metadata
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_INT32 != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not Int32!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_int32);
						track = deadbeef->pl_get_for_idx(a_int32);
						mpris_metadata(&ret, track);
						if (track) {
							deadbeef->pl_item_unref(track);
						}
					}
				} else if (dbus_message_has_member(msg, "GetCurrentTrack")) {
					// Return the position of current URI in the TrackList The return value is zero-based, 
					// so the position of the first URI in the TrackList is 0. The behavior of this method 
					// is unspecified if there are zero elements in the TrackList.
					a_int32 = deadbeef->pl_get_cursor(PL_MAIN);
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_INT32, &a_int32);
				} else if (dbus_message_has_member(msg, "GetLength")) {
					// Returns the number of elements in the TrackList. 
					a_int32 = deadbeef->pl_getcount(PL_MAIN);
					dbus_message_iter_append_basic(&ret, DBUS_TYPE_INT32, &a_int32);
				} else if (dbus_message_has_member(msg, "AddTrack")) {
					// Appends an URI in the TrackList. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_STRING != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not String!\n");
					} else {
						char *file;
						dbus_message_iter_get_basic(&args, &file);
						if(!dbus_message_iter_next(&args)) {
							fprintf(stderr, "mpris: Message has only one arguments!\n");
						} else if (DBUS_TYPE_BOOLEAN != dbus_message_iter_get_arg_type(&args)) {
							fprintf(stderr, "mpris: Argument is not Boolean!\n");
						} else {
							dbus_message_iter_get_basic(&args, &a_bool);
							deadbeef->pl_add_files_begin();
							deadbeef->pl_add_file(file, NULL, NULL);
							deadbeef->pl_add_files_end();
							deadbeef->sendmessage(M_PLAYLISTREFRESH, 0, 0, 0);
							if (a_bool) {
								// PlayImmediately
								track = deadbeef->pl_get_last(PL_MAIN);
								deadbeef->pl_playqueue_push(track);
								deadbeef->playback_next();
								deadbeef->pl_item_unref(track);
							}
						}
					}
				} else if (dbus_message_has_member(msg, "DelTrack")) {
					// Removes an URI from the TrackList by its position in the tracklist
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_INT32 != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not Int32!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_int32);
						track = deadbeef->pl_get_for_idx(a_int32);
						if (track) {
							// FIXME: any better way?
							if (deadbeef->pl_getselcount() > 0) {
								DB_playItem_t *item = deadbeef->pl_get_first(PL_MAIN);
								while(item) {
									deadbeef->pl_set_selected(item, 0);
									deadbeef->pl_item_unref(item);
									item = deadbeef->pl_get_next(item, PL_MAIN);
								}
							}
							deadbeef->pl_set_selected(track, 1);
							deadbeef->pl_delete_selected();
							deadbeef->sendmessage(M_PLAYLISTREFRESH, 0, 0, 0);
						}
					}
				} else if (dbus_message_has_member(msg, "SetLoop")) {
					// Toggle playlist loop. TRUE to loop, FALSE to stop looping. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_BOOLEAN != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not boolean!\n");
					} else {
						dbus_message_iter_get_basic(&args, &a_bool);
						deadbeef->conf_set_int ("playback.loop", a_bool?PLAYBACK_MODE_LOOP_ALL:PLAYBACK_MODE_NOLOOP);
						deadbeef->sendmessage(M_CONFIGCHANGED, 0, 0, 0);
					}
				} else if (dbus_message_has_member(msg, "SetRandom")) {
					// Toggle playlist shuffle / random. It may or may not play tracks only once.
					// TRUE to play randomly / shuffle playlist, FALSE to play normally / reorder playlist. 
					if (!dbus_message_iter_init(msg, &args)) {
						fprintf(stderr, "mpris: Message has no arguments!\n");
					} else if (DBUS_TYPE_BOOLEAN != dbus_message_iter_get_arg_type(&args)) {
						fprintf(stderr, "mpris: Argument is not boolean!\n");
					} else {
						dbus_message_iter_get_basic (&args, &a_bool);
						deadbeef->conf_set_int ("playback.order", a_bool?PLAYBACK_ORDER_RANDOM:PLAYBACK_ORDER_LINEAR);
						deadbeef->sendmessage (M_CONFIGCHANGED, 0, 0, 0);
					}
				}
			}
		}
		
		// send the reply
		dbus_connection_send (conn, reply, NULL);
		dbus_connection_flush (conn);

		// free the reply
		dbus_message_unref (reply);

		// free the message
		dbus_message_unref (msg);
	}
}

int mpris_tracklist_signals (DB_event_t *ev, uintptr_t data) {
	DBusConnection *conn;
	DBusMessage *signal = NULL;
	DBusMessageIter args;
	dbus_uint32_t tracks;

	// printf("tracklist got %d\n", (int)data);
	
	conn = dbus_bus_get(DBUS_BUS_SESSION, NULL);
	if (!conn) {
		return 0;
	}

	// Signal is emitted when the "TrackList" content has changed:
    //  When one or more elements have been added
    //  When one or more elements have been removed  //TODO
    //  When the ordering of elements has changed  //TODO
	// Returns the number of elements in the TrackList after the change happened. 
	signal = dbus_message_new_signal (MPRIS_PATH_TRACKLIST, MPRIS_INTERFACE, "TrackListChange");

	dbus_message_iter_init_append (signal, &args);
	tracks = deadbeef->pl_getcount (PL_MAIN);
	dbus_message_iter_append_basic (&args, DBUS_TYPE_INT32, &tracks);
	
	// send the signal
	dbus_connection_send (conn, signal, NULL);
	dbus_connection_flush (conn);
	
	// free the signal
	dbus_message_unref (signal);
	dbus_connection_unref (conn);
	return 0;
}

int mpris_player_signals (DB_event_t *ev, uintptr_t data) {
	DBusConnection *conn;
	DBusMessage *signal = NULL;
	DBusMessageIter args;
	DBusPendingCall *reply;
	dbus_uint32_t caps;

	// printf ("player got %d\n", (int)data);

	conn = dbus_bus_get(DBUS_BUS_SESSION, NULL);
	if (!conn) {
		return 0;
	}

	// TrackChange Signal is emitted when the "Media Player" plays another "Track". Argument of the signal is the metadata attached to the new "Track".
	if (data == DB_EV_SONGSTARTED) {
		signal = dbus_message_new_signal (MPRIS_PATH_PLAYER, MPRIS_INTERFACE, "TrackChange");
		dbus_message_iter_init_append (signal, &args);
		DB_playItem_t *track = deadbeef->streamer_get_playing_track ();
		mpris_metadata (&args, track);
		dbus_connection_send (conn, signal, NULL);
		dbus_connection_flush (conn);
		deadbeef->pl_item_unref(track);
		dbus_message_unref (signal);
		// printf("player send TrackChange\n");
	}
	
	// CapsChange Signal is emitted when the "Media Player" changes capabilities, see GetCaps method.
	if (data != DB_EV_CONFIGCHANGED) {
		caps = mpris_caps(data);
		if (db_conf[2] != caps) {
			db_conf[2] = caps;
			signal = dbus_message_new_signal (MPRIS_PATH_PLAYER, MPRIS_INTERFACE, "CapsChange");
			dbus_message_iter_init_append (signal, &args);
			dbus_message_iter_append_basic (&args, DBUS_TYPE_INT32, &caps);
			dbus_connection_send (conn, signal, NULL);
			dbus_connection_flush (conn);
			dbus_message_unref (signal);
			// printf ("player send CapsChange\n");
		}
	}

	// StatusChange Signal is emitted when the status of the "Media Player" change. The argument has the same meaning as the value returned by GetStatus.
	if (data == DB_EV_CONFIGCHANGED) {
		int c1 = deadbeef->conf_get_int ("playback.order", 0);
		int c2 = deadbeef->conf_get_int ("playback.loop", 0);
		if (c1 == db_conf[0] && c2 == db_conf[1]) {
			// nothing changed
			return 0;
		}
		db_conf[0] = c1;
		db_conf[1] = c2;
	}
	signal = dbus_message_new_signal (MPRIS_PATH_PLAYER, MPRIS_INTERFACE, "StatusChange");
	dbus_message_iter_init_append (signal, &args);
	// even on DB_EV_SONGSTARTED/DB_EV_SONGFINISHED, deadbeef->get_output()->state() returns OUTPUT_STATE_STOPPED/OUTPUT_STATE_PLAYING
	mpris_status (&args, data);
	dbus_connection_send (conn, signal, NULL);
	dbus_connection_flush (conn);
	dbus_message_unref (signal);
	// printf ("player send StatusChange\n");

	dbus_connection_unref (conn);
	return 0;
}

int mpris_start (void) {
	pthread_t mpris_tid;

	dbus_loop = 1;

	db_conf[0] = deadbeef->conf_get_int ("playback.order", 0);
	db_conf[1] = deadbeef->conf_get_int ("playback.loop", 0);
	db_conf[2] = mpris_caps(-1);

    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_SONGSTARTED, DB_CALLBACK (mpris_player_signals), DB_EV_SONGSTARTED);
    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_SONGFINISHED, DB_CALLBACK (mpris_player_signals), DB_EV_SONGFINISHED);
    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_CONFIGCHANGED, DB_CALLBACK (mpris_player_signals), DB_EV_CONFIGCHANGED);
    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_PAUSED, DB_CALLBACK (mpris_player_signals), DB_EV_PAUSED);
    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_PLAYLISTCHANGED, DB_CALLBACK (mpris_tracklist_signals), DB_EV_PLAYLISTCHANGED);
    deadbeef->ev_subscribe (DB_PLUGIN (&plugin), DB_EV_PLAYLISTSWITCH, DB_CALLBACK (mpris_tracklist_signals), DB_EV_PLAYLISTSWITCH);

	pthread_create (&mpris_tid, NULL, mpris, NULL);
    return 0;
}

int mpris_stop (void) {
	dbus_loop = 0;

    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_SONGSTARTED, DB_CALLBACK (mpris_player_signals), DB_EV_SONGSTARTED);
    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_SONGFINISHED, DB_CALLBACK (mpris_player_signals), DB_EV_SONGFINISHED);
    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_CONFIGCHANGED, DB_CALLBACK (mpris_player_signals), DB_EV_CONFIGCHANGED);
    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_PAUSED, DB_CALLBACK (mpris_player_signals), DB_EV_PAUSED);
    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_PLAYLISTCHANGED, DB_CALLBACK (mpris_tracklist_signals), DB_EV_PLAYLISTCHANGED);
    deadbeef->ev_unsubscribe (DB_PLUGIN (&plugin), DB_EV_PLAYLISTSWITCH, DB_CALLBACK (mpris_tracklist_signals), DB_EV_PLAYLISTSWITCH);
    return 0;
}

DB_misc_t plugin = {
    DB_PLUGIN_SET_API_VERSION
    .plugin.type = DB_PLUGIN_MISC,
    .plugin.version_major = 0,
    .plugin.version_minor = 9,
    .plugin.id = "mpris",
    .plugin.name = "MPRIS",
    .plugin.descr = "Support Media Player Remote Interfacing Specification v1.0",
    .plugin.author = "xuzhen",
    .plugin.email = "xuzhen666@gmail.com",
    .plugin.website = "http://deadbeef.sourceforge.net",
    .plugin.start = mpris_start,
    .plugin.stop = mpris_stop,
};

DB_plugin_t *
mpris_load (DB_functions_t *ddb) {
    deadbeef = ddb;
    return &plugin.plugin;
}
